@@ -4,57 +4,70 @@ use serde::Serialize;
 use cosmwasm_std::testing::{mock_env, MockApi};
 use cosmwasm_std::{
     from_slice, to_binary, to_vec, Addr, Api, Attribute, BankMsg, Binary, BlockInfo, Coin,
-    ContractResult, CosmosMsg, Empty, MessageInfo, Querier, QuerierResult, QuerierWrapper,
-    QueryRequest, Response, SystemError, SystemResult, WasmMsg,
+    ContractResult, CosmosMsg, Empty, Event, MessageInfo, Querier, QuerierResult, QuerierWrapper,
+    QueryRequest, Reply, ReplyOn, Response, SubMsg, SubMsgExecutionResponse, SystemError,
+    SystemResult, WasmMsg,
 };
 
-use crate::bank::{Bank, BankCache, BankOps, BankRouter};
-use crate::wasm::{Contract, StorageFactory, WasmCache, WasmOps, WasmRouter};
+use crate::bank::{Bank, BankCache, BankCommittable, BankOps, BankRouter};
+use crate::wasm::{Contract, StorageFactory, WasmCache, WasmCommittable, WasmOps, WasmRouter};
 use schemars::JsonSchema;
 use std::fmt;
 
+use prost::Message;
+// use bytes::buf::BufMut;
+
 #[derive(Default, Clone, Debug)]
 pub struct AppResponse {
-    pub attributes: Vec<Attribute>,
+    pub events: Vec<Event>,
     pub data: Option<Binary>,
 }
 
-// This can be Response, Response, MigrationResponse
-#[derive(Default, Clone)]
-pub struct ActionResponse<C>
-where
-    C: Clone + fmt::Debug + PartialEq + JsonSchema,
-{
-    // TODO: allow T != Empty
-    pub messages: Vec<CosmosMsg<C>>,
-    pub attributes: Vec<Attribute>,
-    pub data: Option<Binary>,
-}
-
-impl<C> From<Response<C>> for ActionResponse<C>
-where
-    C: Clone + fmt::Debug + PartialEq + JsonSchema,
-{
-    fn from(input: Response<C>) -> Self {
-        ActionResponse {
-            messages: input.messages,
-            attributes: input.attributes,
-            data: input.data,
-        }
+impl AppResponse {
+    // Return all custom attributes returned by the contract in the `idx` event.
+    // We assert the type is wasm, and skip the contract_address attribute.
+    pub fn custom_attrs(&self, idx: usize) -> &[Attribute] {
+        assert_eq!(self.events[idx].ty.as_str(), "wasm");
+        &self.events[idx].attributes[1..]
     }
 }
 
-impl<C> ActionResponse<C>
+#[derive(Clone, PartialEq, Message)]
+pub struct InstantiateData {
+    #[prost(string, tag = "1")]
+    pub address: ::prost::alloc::string::String,
+    /// Unique ID number for this person.
+    #[prost(bytes, tag = "2")]
+    pub data: ::prost::alloc::vec::Vec<u8>,
+}
+
+fn init_response<C>(res: &mut Response<C>, contact_address: &Addr)
 where
     C: Clone + fmt::Debug + PartialEq + JsonSchema,
 {
-    fn init(input: Response<C>, address: Addr) -> Self {
-        ActionResponse {
-            messages: input.messages,
-            attributes: input.attributes,
-            data: Some(address.as_ref().as_bytes().into()),
-        }
+    let data = res.data.clone().unwrap_or_default().to_vec();
+    let init_data = InstantiateData {
+        address: contact_address.into(),
+        data,
+    };
+    let mut new_data = Vec::<u8>::with_capacity(init_data.encoded_len());
+    // the data must encode successfully
+    init_data.encode(&mut new_data).unwrap();
+    res.data = Some(new_data.into());
+}
+
+// this parses the result from a wasm contract init
+pub fn parse_contract_addr(data: &Option<Binary>) -> Result<Addr, String> {
+    let bin = data
+        .as_ref()
+        .ok_or_else(|| "No data response".to_string())?
+        .to_vec();
+    // parse the protobuf struct
+    let init_data = InstantiateData::decode(bin.as_slice()).map_err(|e| e.to_string())?;
+    if init_data.address.is_empty() {
+        return Err("no contract address provided".into());
     }
+    Ok(Addr::unchecked(init_data.address))
 }
 
 impl<C> Querier for App<C>
@@ -165,15 +178,14 @@ where
     ) -> Result<Addr, String> {
         // instantiate contract
         let init_msg = to_binary(init_msg).map_err(|e| e.to_string())?;
-        let msg: CosmosMsg<C> = WasmMsg::Instantiate {
+        let msg = WasmMsg::Instantiate {
             admin: None,
             code_id,
             msg: init_msg,
-            send: send_funds.to_vec(),
+            funds: send_funds.to_vec(),
             label: label.into(),
-        }
-        .into();
-        let res = self.execute(sender, msg)?;
+        };
+        let res = self.execute(sender, msg.into())?;
         parse_contract_addr(&res.data)
     }
 
@@ -190,10 +202,9 @@ where
         let msg = WasmMsg::Execute {
             contract_addr: contract_addr.into(),
             msg,
-            send: send_funds.to_vec(),
-        }
-        .into();
-        self.execute(sender, msg)
+            funds: send_funds.to_vec(),
+        };
+        self.execute(sender, msg.into())
     }
 
     /// Runs arbitrary CosmosMsg.
@@ -259,23 +270,49 @@ pub struct AppCache<'a, C>
 where
     C: Clone + fmt::Debug + PartialEq + JsonSchema,
 {
-    router: &'a App<C>,
+    querier: &'a dyn Querier,
     wasm: WasmCache<'a, C>,
     bank: BankCache<'a>,
 }
 
+pub trait AppCommittable {
+    fn commit_bank(&mut self) -> &mut dyn BankCommittable;
+    fn commit_wasm(&mut self) -> &mut dyn WasmCommittable;
+}
+
+impl<'a, C> AppCommittable for AppCache<'a, C>
+where
+    C: Clone + fmt::Debug + PartialEq + JsonSchema,
+{
+    fn commit_bank(&mut self) -> &mut dyn BankCommittable {
+        &mut self.bank
+    }
+    fn commit_wasm(&mut self) -> &mut dyn WasmCommittable {
+        &mut self.wasm
+    }
+}
+
+impl<C> AppCommittable for App<C>
+where
+    C: Clone + fmt::Debug + PartialEq + JsonSchema,
+{
+    fn commit_bank(&mut self) -> &mut dyn BankCommittable {
+        &mut self.bank
+    }
+    fn commit_wasm(&mut self) -> &mut dyn WasmCommittable {
+        &mut self.wasm
+    }
+}
+
 pub struct AppOps {
     wasm: WasmOps,
     bank: BankOps,
 }
 
 impl AppOps {
-    pub fn commit<C>(self, router: &mut App<C>)
-    where
-        C: Clone + fmt::Debug + PartialEq + JsonSchema,
-    {
-        self.bank.commit(&mut router.bank);
-        self.wasm.commit(&mut router.wasm);
+    pub fn commit(self, commit: &mut dyn AppCommittable) {
+        self.bank.commit(commit.commit_bank());
+        self.wasm.commit(commit.commit_wasm());
     }
 }
 
@@ -285,12 +322,20 @@ where
 {
     fn new(router: &'a App<C>) -> Self {
         AppCache {
-            router,
+            querier: router,
             wasm: router.wasm.cache(),
             bank: router.bank.cache(),
         }
     }
 
+    pub fn cache(&self) -> AppCache<C> {
+        AppCache {
+            querier: self.querier,
+            wasm: self.wasm.cache(),
+            bank: self.bank.cache(),
+        }
+    }
+
     /// When we want to commit the RouterCache, we need a 2 step process to satisfy Rust reference counting:
     /// 1. prepare() consumes RouterCache, releasing &Router, and creating a self-owned update info.
     /// 2. RouterOps::commit() can now take &mut Router and updates the underlying state
@@ -301,28 +346,11 @@ where
         }
     }
 
-    /// This will execute the given messages, making all changes to the local cache.
-    /// This *will* write some data to the cache if the message fails half-way through.
-    /// All sequential calls to RouterCache will be one atomic unit (all commit or all fail).
-    ///
-    /// For normal use cases, you can use Router::execute() or Router::execute_multi().
-    /// This is designed to be handled internally as part of larger process flows.
-    fn execute(&mut self, sender: Addr, msg: CosmosMsg<C>) -> Result<AppResponse, String> {
+    pub fn execute(&mut self, sender: Addr, msg: CosmosMsg<C>) -> Result<AppResponse, String> {
         match msg {
             CosmosMsg::Wasm(msg) => {
-                let (resender, res) = self.handle_wasm(sender, msg)?;
-                let mut attributes = res.attributes;
-                // recurse in all messages
-                for resend in res.messages {
-                    let subres = self.execute(resender.clone(), resend)?;
-                    // ignore the data now, just like in wasmd
-                    // append the events
-                    attributes.extend_from_slice(&subres.attributes);
-                }
-                Ok(AppResponse {
-                    attributes,
-                    data: res.data,
-                })
+                let (resender, res) = self.execute_wasm(sender, msg)?;
+                self.process_response(resender, res, false)
             }
             CosmosMsg::Bank(msg) => {
                 self.bank.execute(sender, msg)?;
@@ -332,69 +360,144 @@ where
         }
     }
 
-    fn sudo(&mut self, contract_addr: Addr, msg: Vec<u8>) -> Result<AppResponse, String> {
-        let res = self.wasm.sudo(contract_addr.clone(), self.router, msg)?;
-        let mut attributes = res.attributes;
+    /// This will execute the given messages, making all changes to the local cache.
+    /// This *will* write some data to the cache if the message fails half-way through.
+    /// All sequential calls to RouterCache will be one atomic unit (all commit or all fail).
+    ///
+    /// For normal use cases, you can use Router::execute() or Router::execute_multi().
+    /// This is designed to be handled internally as part of larger process flows.
+    fn execute_submsg(&mut self, contract: Addr, msg: SubMsg<C>) -> Result<AppResponse, String> {
+        // execute in cache
+        let mut subtx = self.cache();
+        let res = subtx.execute(contract.clone(), msg.msg);
+        if res.is_ok() {
+            subtx.prepare().commit(self);
+        }
+
+        // call reply if meaningful
+        if let Ok(r) = res {
+            if matches!(msg.reply_on, ReplyOn::Always | ReplyOn::Success) {
+                let mut orig = r.clone();
+                let reply = Reply {
+                    id: msg.id,
+                    result: ContractResult::Ok(SubMsgExecutionResponse {
+                        events: r.events,
+                        data: r.data,
+                    }),
+                };
+                // do reply and combine it with the original response
+                let res2 = self._reply(contract, reply)?;
+                // override data if set
+                if let Some(data) = res2.data {
+                    orig.data = Some(data);
+                }
+                // append the events
+                orig.events.extend_from_slice(&res2.events);
+                Ok(orig)
+            } else {
+                Ok(r)
+            }
+        } else if let Err(e) = res {
+            if matches!(msg.reply_on, ReplyOn::Always | ReplyOn::Error) {
+                let reply = Reply {
+                    id: msg.id,
+                    result: ContractResult::Err(e),
+                };
+                self._reply(contract, reply)
+            } else {
+                Err(e)
+            }
+        } else {
+            res
+        }
+    }
+
+    fn _reply(&mut self, contract: Addr, reply: Reply) -> Result<AppResponse, String> {
+        // TODO: process result better, combine events / data from parent
+        let res = self.wasm.reply(contract.clone(), self.querier, reply)?;
+        self.process_response(contract, res, true)
+    }
+
+    fn process_response(
+        &mut self,
+        contract: Addr,
+        response: Response<C>,
+        ignore_attributes: bool,
+    ) -> Result<AppResponse, String> {
+        // These need to get `wasm-` prefix to match the wasmd semantics (custom wasm messages cannot
+        // fake system level event types, like transfer from the bank module)
+        let mut events: Vec<_> = response
+            .events
+            .into_iter()
+            .map(|mut ev| {
+                ev.ty = format!("wasm-{}", ev.ty);
+                ev
+            })
+            .collect();
+        // hmmm... we don't need this for reply, right?
+        if !ignore_attributes {
+            // turn attributes into event and place it first
+            let mut wasm_event = Event::new("wasm").attr("contract_address", &contract);
+            wasm_event
+                .attributes
+                .extend_from_slice(&response.attributes);
+            events.insert(0, wasm_event);
+        }
+
         // recurse in all messages
-        for resend in res.messages {
-            let subres = self.execute(contract_addr.clone(), resend)?;
-            // ignore the data now, just like in wasmd
-            // append the events
-            attributes.extend_from_slice(&subres.attributes);
+        for resend in response.messages {
+            let subres = self.execute_submsg(contract.clone(), resend)?;
+            events.extend_from_slice(&subres.events);
         }
         Ok(AppResponse {
-            attributes,
-            data: res.data,
+            events,
+            data: response.data,
         })
     }
 
+    fn sudo(&mut self, contract_addr: Addr, msg: Vec<u8>) -> Result<AppResponse, String> {
+        let res = self.wasm.sudo(contract_addr.clone(), self.querier, msg)?;
+        self.process_response(contract_addr, res, false)
+    }
+
     // this returns the contract address as well, so we can properly resend the data
-    fn handle_wasm(
-        &mut self,
-        sender: Addr,
-        msg: WasmMsg,
-    ) -> Result<(Addr, ActionResponse<C>), String> {
+    fn execute_wasm(&mut self, sender: Addr, msg: WasmMsg) -> Result<(Addr, Response<C>), String> {
         match msg {
             WasmMsg::Execute {
                 contract_addr,
                 msg,
-                send,
+                funds,
             } => {
                 let contract_addr = Addr::unchecked(contract_addr);
                 // first move the cash
-                self.send(sender.clone(), contract_addr.clone().into(), &send)?;
+                self.send(sender.clone(), contract_addr.clone().into(), &funds)?;
                 // then call the contract
-                let info = MessageInfo {
-                    sender,
-                    funds: send,
-                };
+                let info = MessageInfo { sender, funds };
                 let res =
                     self.wasm
-                        .handle(contract_addr.clone(), self.router, info, msg.to_vec())?;
-                Ok((contract_addr, res.into()))
+                        .execute(contract_addr.clone(), self.querier, info, msg.to_vec())?;
+                Ok((contract_addr, res))
             }
             WasmMsg::Instantiate {
                 admin: _,
                 code_id,
                 msg,
-                send,
+                funds,
                 label: _,
             } => {
                 let contract_addr = Addr::unchecked(self.wasm.register_contract(code_id as usize)?);
                 // move the cash
-                self.send(sender.clone(), contract_addr.clone().into(), &send)?;
+                self.send(sender.clone(), contract_addr.clone().into(), &funds)?;
                 // then call the contract
-                let info = MessageInfo {
-                    sender,
-                    funds: send,
-                };
-                let res = self
-                    .wasm
-                    .init(contract_addr.clone(), self.router, info, msg.to_vec())?;
-                Ok((
+                let info = MessageInfo { sender, funds };
+                let mut res = self.wasm.instantiate(
                     contract_addr.clone(),
-                    ActionResponse::init(res, contract_addr),
-                ))
+                    self.querier,
+                    info,
+                    msg.to_vec(),
+                )?;
+                init_response(&mut res, &contract_addr);
+                Ok((contract_addr, res))
             }
             WasmMsg::Migrate { .. } => unimplemented!(),
             m => panic!("Unsupported wasm message: {:?}", m),
@@ -418,26 +521,17 @@ where
     }
 }
 
-// this parses the result from a wasm contract init
-pub fn parse_contract_addr(data: &Option<Binary>) -> Result<Addr, String> {
-    let bin = data
-        .as_ref()
-        .ok_or_else(|| "No data response".to_string())?
-        .to_vec();
-    let str = String::from_utf8(bin).map_err(|e| e.to_string())?;
-    Ok(Addr::unchecked(str))
-}
-
 #[cfg(test)]
 mod test {
     use super::*;
     use crate::test_helpers::{
         contract_payout, contract_payout_custom, contract_reflect, CustomMsg, EmptyMsg,
-        PayoutMessage, ReflectMessage, ReflectResponse, ReflectSudoMsg,
+        PayoutCountResponse, PayoutInitMessage, PayoutQueryMsg, PayoutSudoMsg, ReflectMessage,
+        ReflectQueryMsg,
     };
     use crate::SimpleBank;
     use cosmwasm_std::testing::MockStorage;
-    use cosmwasm_std::{attr, coin, coins};
+    use cosmwasm_std::{attr, coin, coins, Reply};
 
     fn mock_router() -> App {
         let env = mock_env();
@@ -514,7 +608,7 @@ mod test {
 
         // set up contract
         let code_id = router.store_code(contract_payout());
-        let msg = PayoutMessage {
+        let msg = PayoutInitMessage {
             payout: coin(5, "eth"),
         };
         let contract_addr = router
@@ -537,8 +631,9 @@ mod test {
         let res = router
             .execute_contract(random.clone(), contract_addr.clone(), &EmptyMsg {}, &[])
             .unwrap();
-        assert_eq!(1, res.attributes.len());
-        assert_eq!(&attr("action", "payout"), &res.attributes[0]);
+        assert_eq!(1, res.events.len());
+        let custom_attrs = res.custom_attrs(0);
+        assert_eq!(&[attr("action", "payout")], &custom_attrs);
 
         // random got cash
         let funds = get_balance(&router, &random);
@@ -559,7 +654,7 @@ mod test {
 
         // set up payout contract
         let payout_id = router.store_code(contract_payout_custom());
-        let msg = PayoutMessage {
+        let msg = PayoutInitMessage {
             payout: coin(5, "eth"),
         };
         let payout_addr = router
@@ -576,19 +671,18 @@ mod test {
         let funds = get_balance(&router, &reflect_addr);
         assert_eq!(funds, vec![]);
         // reflect count is 1
-        let qres: ReflectResponse = router
+        let qres: PayoutCountResponse = router
             .wrap()
-            .query_wasm_smart(&reflect_addr, &EmptyMsg {})
+            .query_wasm_smart(&reflect_addr, &ReflectQueryMsg::Count {})
             .unwrap();
-        assert_eq!(1, qres.count);
+        assert_eq!(0, qres.count);
 
         // reflecting payout message pays reflect contract
-        let msg = WasmMsg::Execute {
-            contract_addr: payout_addr.into(),
+        let msg = SubMsg::new(WasmMsg::Execute {
+            contract_addr: payout_addr.clone().into(),
             msg: b"{}".into(),
-            send: vec![],
-        }
-        .into();
+            funds: vec![],
+        });
         let msgs = ReflectMessage {
             messages: vec![msg],
         };
@@ -597,19 +691,36 @@ mod test {
             .unwrap();
 
         // ensure the attributes were relayed from the sub-message
-        assert_eq!(1, res.attributes.len());
-        assert_eq!(&attr("action", "payout"), &res.attributes[0]);
+        assert_eq!(2, res.events.len(), "{:?}", res.events);
+        // first event was the call to reflect
+        let first = &res.events[0];
+        assert_eq!(first.ty.as_str(), "wasm");
+        assert_eq!(1, first.attributes.len());
+        assert_eq!(
+            &attr("contract_address", &reflect_addr),
+            &first.attributes[0]
+        );
+        // second event was call to payout
+        let second = &res.events[1];
+        assert_eq!(second.ty.as_str(), "wasm");
+        assert_eq!(2, second.attributes.len());
+        assert_eq!(
+            &attr("contract_address", &payout_addr),
+            &second.attributes[0]
+        );
+        assert_eq!(&attr("action", "payout"), &second.attributes[1]);
+        // FIXME? reply didn't add any more events itself...
 
         // ensure transfer was executed with reflect as sender
         let funds = get_balance(&router, &reflect_addr);
         assert_eq!(funds, coins(5, "eth"));
 
         // reflect count updated
-        let qres: ReflectResponse = router
+        let qres: PayoutCountResponse = router
             .wrap()
-            .query_wasm_smart(&reflect_addr, &EmptyMsg {})
+            .query_wasm_smart(&reflect_addr, &ReflectQueryMsg::Count {})
             .unwrap();
-        assert_eq!(2, qres.count);
+        assert_eq!(1, qres.count);
     }
 
     #[test]
@@ -639,40 +750,41 @@ mod test {
         let random = Addr::unchecked("random");
 
         // sending 7 eth works
-        let msg = BankMsg::Send {
+        let msg = SubMsg::new(BankMsg::Send {
             to_address: random.clone().into(),
             amount: coins(7, "eth"),
-        }
-        .into();
+        });
         let msgs = ReflectMessage {
             messages: vec![msg],
         };
         let res = router
             .execute_contract(random.clone(), reflect_addr.clone(), &msgs, &[])
             .unwrap();
-        assert_eq!(0, res.attributes.len());
+        // only one wasm event with no custom attributes
+        assert_eq!(1, res.events.len());
+        assert_eq!(1, res.events[0].attributes.len());
+        assert_eq!("wasm", res.events[0].ty.as_str());
+        assert_eq!("contract_address", res.events[0].attributes[0].key.as_str());
         // ensure random got paid
         let funds = get_balance(&router, &random);
         assert_eq!(funds, coins(7, "eth"));
 
-        // reflect count should be updated to 2
-        let qres: ReflectResponse = router
+        // reflect count should be updated to 1
+        let qres: PayoutCountResponse = router
             .wrap()
-            .query_wasm_smart(&reflect_addr, &EmptyMsg {})
+            .query_wasm_smart(&reflect_addr, &ReflectQueryMsg::Count {})
             .unwrap();
-        assert_eq!(2, qres.count);
+        assert_eq!(1, qres.count);
 
         // sending 8 eth, then 3 btc should fail both
-        let msg = BankMsg::Send {
+        let msg = SubMsg::new(BankMsg::Send {
             to_address: random.clone().into(),
             amount: coins(8, "eth"),
-        }
-        .into();
-        let msg2 = BankMsg::Send {
+        });
+        let msg2 = SubMsg::new(BankMsg::Send {
             to_address: random.clone().into(),
             amount: coins(3, "btc"),
-        }
-        .into();
+        });
         let msgs = ReflectMessage {
             messages: vec![msg, msg2],
         };
@@ -686,39 +798,134 @@ mod test {
         assert_eq!(funds, coins(7, "eth"));
 
         // failure should not update reflect count
-        let qres: ReflectResponse = router
+        let qres: PayoutCountResponse = router
             .wrap()
-            .query_wasm_smart(&reflect_addr, &EmptyMsg {})
+            .query_wasm_smart(&reflect_addr, &ReflectQueryMsg::Count {})
             .unwrap();
-        assert_eq!(2, qres.count);
+        assert_eq!(1, qres.count);
     }
 
     #[test]
     fn sudo_works() {
-        let mut router = custom_router();
+        let mut router = mock_router();
 
         let owner = Addr::unchecked("owner");
-        let reflect_id = router.store_code(contract_reflect());
-        let reflect_addr = router
-            .instantiate_contract(reflect_id, owner, &EmptyMsg {}, &[], "Reflect")
+        let init_funds = vec![coin(100, "eth")];
+        router.set_bank_balance(&owner, init_funds).unwrap();
+        let payout_id = router.store_code(contract_payout());
+        let msg = PayoutInitMessage {
+            payout: coin(5, "eth"),
+        };
+        let payout_addr = router
+            .instantiate_contract(payout_id, owner, &msg, &coins(23, "eth"), "Payout")
             .unwrap();
 
         // count is 1
-        let ReflectResponse { count } = router
+        let PayoutCountResponse { count } = router
             .wrap()
-            .query_wasm_smart(&reflect_addr, &EmptyMsg {})
+            .query_wasm_smart(&payout_addr, &PayoutQueryMsg::Count {})
             .unwrap();
         assert_eq!(1, count);
 
         // sudo
-        let msg = ReflectSudoMsg { set_count: 25 };
-        router.sudo(reflect_addr.clone(), &msg).unwrap();
+        let msg = PayoutSudoMsg { set_count: 25 };
+        router.sudo(payout_addr.clone(), &msg).unwrap();
 
         // count is 25
-        let ReflectResponse { count } = router
+        let PayoutCountResponse { count } = router
             .wrap()
-            .query_wasm_smart(&reflect_addr, &EmptyMsg {})
+            .query_wasm_smart(&payout_addr, &PayoutQueryMsg::Count {})
             .unwrap();
         assert_eq!(25, count);
     }
+
+    #[test]
+    fn reflect_submessage_reply_works() {
+        let mut router = custom_router();
+
+        // set personal balance
+        let owner = Addr::unchecked("owner");
+        let random = Addr::unchecked("random");
+        let init_funds = vec![coin(20, "btc"), coin(100, "eth")];
+        router.set_bank_balance(&owner, init_funds).unwrap();
+
+        // set up reflect contract
+        let reflect_id = router.store_code(contract_reflect());
+        let reflect_addr = router
+            .instantiate_contract(
+                reflect_id,
+                owner,
+                &EmptyMsg {},
+                &coins(40, "eth"),
+                "Reflect",
+            )
+            .unwrap();
+
+        // no reply writen beforehand
+        let query = ReflectQueryMsg::Reply { id: 123 };
+        router
+            .wrap()
+            .query_wasm_smart::<Reply, _, _>(&reflect_addr, &query)
+            .unwrap_err();
+
+        // reflect sends 7 eth, success
+        let msg = SubMsg::reply_always(
+            BankMsg::Send {
+                to_address: random.clone().into(),
+                amount: coins(7, "eth"),
+            },
+            123,
+        );
+        let msgs = ReflectMessage {
+            messages: vec![msg],
+        };
+        let res = router
+            .execute_contract(random.clone(), reflect_addr.clone(), &msgs, &[])
+            .unwrap();
+        // we should get 2 events, the wasm one and the custom event
+        assert_eq!(2, res.events.len(), "{:?}", res.events);
+        // the first one is just the standard wasm message with custom_address (no more attrs)
+        let attrs = res.custom_attrs(0);
+        assert_eq!(0, attrs.len());
+        // the second one is a custom event
+        let second = &res.events[1];
+        assert_eq!("wasm-custom", second.ty.as_str());
+        assert_eq!(2, second.attributes.len());
+        assert_eq!(&attr("from", "reply"), &second.attributes[0]);
+        assert_eq!(&attr("to", "test"), &second.attributes[1]);
+
+        // ensure success was written
+        let res: Reply = router
+            .wrap()
+            .query_wasm_smart(&reflect_addr, &query)
+            .unwrap();
+        assert_eq!(res.id, 123);
+        assert!(res.result.is_ok());
+        // TODO: any more checks on reply data???
+
+        // reflect sends 300 btc, failure, but error caught by submessage (so shows success)
+        let msg = SubMsg::reply_always(
+            BankMsg::Send {
+                to_address: random.clone().into(),
+                amount: coins(300, "btc"),
+            },
+            456,
+        );
+        let msgs = ReflectMessage {
+            messages: vec![msg],
+        };
+        let _res = router
+            .execute_contract(random, reflect_addr.clone(), &msgs, &[])
+            .unwrap();
+
+        // ensure error was written
+        let query = ReflectQueryMsg::Reply { id: 456 };
+        let res: Reply = router
+            .wrap()
+            .query_wasm_smart(&reflect_addr, &query)
+            .unwrap();
+        assert_eq!(res.id, 456);
+        assert!(res.result.is_err());
+        // TODO: check error?
+    }
 }
